/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <float.h>
#include <math.h>
#include <stdarg.h>
#include "taucs.h"

/********************************************************************/
/* taucs_system_memory_size                                         */
/*   returns size of memory reported by the operating system        */
/*   should not normally be called by the user (call _avail_)       */
/********************************************************************/

#ifdef OSTYPE_linux
#define TAUCS_SYSTEM_MEMORY_SIZE_DEFINED

double taucs_system_memory_size()
{
  FILE* f;
  double m;
  double m_sc;

  m_sc  = (double) sysconf(_SC_PAGESIZE);
  m_sc *= (double) sysconf(_SC_PHYS_PAGES);

  /* total memory is the first number in /proc/meminfo */

  f = fopen("/proc/meminfo","r");
  if (f==NULL) return m_sc;
  if (fscanf(f,"%*[a-zA-Z :\n\r]%lf",&m) != 1) return m_sc;

  if (m != m_sc) {
    taucs_printf("Warning: /proc/meminfo reports %lfMB of memory while\n",
	       m/1048576.0);
    taucs_printf("         sysconf       reports %lfMB of memory\n",
	       m_sc/1048576.0);
  }

  return m;
}
#endif

#ifdef OSTYPE_aix
#define TAUCS_SYSTEM_MEMORY_SIZE_DEFINED
double taucs_system_memory_size()
{
  FILE* f;
  double m;

  int child_stdout[2];

  pipe(child_stdout);

  if (fork() == 0) {
    char* argv[10];
    int   i = -1;
    argv[i++] = "lsattr";
    argv[i++] = "-E";
    argv[i++] = "-F";
    argv[i++] = "value";
    argv[i++] = "-l";
    argv[i++] = "sys0";
    argv[i++] = "-a";
    argv[i++] = "realmem";
    argv[i++] = 0;
    close(child_stdout[0]);
    dup2(child_stdout[1],1); 
    execv("/usr/sbin/lsattr",argv);
    perror("System error (execv)");
    exit(0);
  } else {
    char buffer[256];
    int  nbytes;
    /* parent continues */
    close(child_stdout[1]);
    nbytes = read(child_stdout[0],buffer,256);
    close(child_stdout[0]);
    if (sscanf(buffer,"%lf",&m) != 1)
      return -1.0;
    return 1024.0 * m;
  }
}
#endif

#ifdef OSTYPE_solaris
#define TAUCS_SYSTEM_MEMORY_SIZE_DEFINED
#include <sys/unistd.h>

double taucs_system_memory_size()
{
  double m;

  m  = (double) sysconf(_SC_PAGESIZE);
  m *= (double) sysconf(_SC_PHYS_PAGES);

  return m;
}
#endif

#ifdef OSTYPE_irix
#define TAUCS_SYSTEM_MEMORY_SIZE_DEFINED
#include <sys/unistd.h>
#include <sys/types.h>
#include <sys/sysmp.h>
#include <sys/sysinfo.h>

double taucs_system_memory_size()
{
  double m;
  struct rminfo p;

  m  = (double) sysconf(_SC_PAGESIZE);
  /*m  = (double) getpagesize();*/

  taucs_printf("***************** %.0lf ************\n",m);

  if (sysmp(MP_SAGET, MPSA_RMINFO, &p) == -1) {
    perror("sysmp");
    exit(1);
  }
  m = (double) (p.physmem);
  taucs_printf("**$$$* %.0lf\n",m);  

  return m;
}
#endif

#ifndef TAUCS_SYSTEM_MEMORY_SIZE_DEFINED
#define TAUCS_SYSTEM_MEMORY_SIZE_DEFINED
double taucs_system_memory_size()
{ 
  taucs_printf("Warning: cannot automatically determine main memory size\n");
  taucs_printf("         for this platform\n");
  return -1.0; 
}
#endif


/********************************************************************/
/* taucs_avail_memory_size                                            */
/*   returns size of memory available for allocation                */
/********************************************************************/

double taucs_available_memory_size() 
{
  double m_sys;
  double m,m_low,m_high,m_tol;
  char*  p;

  m_sys = taucs_system_memory_size();
  
  /* malloc test */

  m = 1048576.0;

  while ( p=(char*) malloc( (size_t) (m*2.0) ) ) {
    free(p);
    m = m*2.0;
  }

  m_low  = m;
  m_high = 2.0*m;
  m_tol  = m / 128.0;

  while ( m_high - m_low > m_tol ) {
    m = m_low + ( (m_high-m_low)/2.0 );
    taucs_printf("taucs_avail_memory_size: [%.0lf %.0lf %.0lf]\n",
	       m_low  / 1048576.0,
	       m      / 1048576.0,
	       m_high / 1048576.0);
    if ( p=(char*) malloc( (size_t) m ) ) 
      m_low = m;
    else 
      m_high = m;
    free(p);
  }

  m = m_low;

  taucs_printf("taucs_avail_memory_size: malloc test=%.0lfMB sys test=%.0lf.MB\n",
	     m / 1048576.0,
	     m_sys / 1048576.0
	     );

  /* if m_sys is meaningful, then we limit m by 0.75*m_sys */

  if (m_sys > 0) {
    m_sys = floor(0.75 * m_sys); 
    if (m_sys < m) m = m_sys;
  }

  return m;
}

